"""Custom doit reporter.

Based on conversation here:
https://groups.google.com/g/python-doit/c/SgoiGt_XYDU/m/PQ8JmlKFAgAJ

"""

from enum import IntEnum

from attrs import field, mutable
from attrs_strict import type_validator
from beartype import beartype
from beartype.typing import Any, List, OrderedDict
from doit.reporter import ConsoleReporter
from doit.task import Task
from sty import fg


class _TaskExitCode(IntEnum):  # noqa: H601
    """Enum for identifying the task exit code."""

    PASS = 0
    FAIL = 1
    NOT_RUN = 2
    SKIP_IGNORED = 3
    SKIP_UP_TO_DATE = 4


@mutable(kw_only=True)
class _TaskSummary:  # noqa: H601
    """Task Summary."""

    name: str = field(validator=type_validator())
    exit_code: _TaskExitCode = field(validator=type_validator())


@beartype
def _format_task_summary(task_summary: _TaskSummary) -> str:
    """Format the string task summary for printing."""
    lookup = {  # noqa: DAR101, DAR201
        _TaskExitCode.PASS: (fg.green, 'was successful'),
        _TaskExitCode.FAIL: (fg.red, 'failed'),
        _TaskExitCode.NOT_RUN: (fg.white, 'was not run'),
        _TaskExitCode.SKIP_IGNORED: (fg.yellow, 'was ignored'),
        _TaskExitCode.SKIP_UP_TO_DATE: (fg.yellow, 'was skipped'),
    }
    foreground, exit_summary = lookup.get(task_summary.exit_code, ('', 'is UNKNOWN'))
    return f'{foreground}{task_summary.name} {exit_summary}{fg.rs}'


class SummaryReporter(ConsoleReporter):  # pragma: no cover # noqa: H601
    """Summarize results of doit tasks."""

    def __init__(self, *args, **kwargs) -> None:
        """Initialize data members."""  # noqa: DAR101
        super().__init__(*args, **kwargs)
        self._all_tasks = []
        self._task_summaries = {}

    def initialize(self, tasks: OrderedDict[str, Task], selected_tasks: List[str]) -> None:
        """Initialize the data members for tracking task run status."""  # noqa: DAR101
        super().initialize(tasks, selected_tasks)
        self._all_tasks = selected_tasks

    def _add_summary(self, task: Task, exit_code: _TaskExitCode) -> None:
        """Store summary for later reference."""  # noqa: DAR101
        task_summary = _TaskSummary(name=task.name, exit_code=exit_code)
        self._task_summaries[task_summary.name] = task_summary

    def add_failure(self, task: Task, exception: Any) -> None:
        """Catch tasks that failed."""  # noqa: DAR101
        super().add_failure(task, exception)
        self._add_summary(task, _TaskExitCode.FAIL)

    def add_success(self, task: Task) -> None:
        """Catch tasks that succeeded."""  # noqa: DAR101
        super().add_success(task)
        self._add_summary(task, _TaskExitCode.PASS)

    def skip_uptodate(self, task: Task) -> None:
        """Catch tasks skipped."""  # noqa: DAR101
        super().skip_uptodate(task)
        self._add_summary(task, _TaskExitCode.SKIP_IGNORED)

    def skip_ignore(self, task: Task) -> None:
        """Catch tasks skipped."""  # noqa: DAR101
        super().skip_ignore(task)
        self._add_summary(task, _TaskExitCode.SKIP_UP_TO_DATE)

    def complete_run(self) -> None:
        """Output a concise summary."""
        super().complete_run()

        # Inspired by "nox"
        prefix = 'doit> '
        self.write(f'\n{prefix}Summary:\n')
        not_run_kwargs = {'exit_code': _TaskExitCode.NOT_RUN}
        for task_name in self._all_tasks:
            task_summary = self._task_summaries.get(task_name, _TaskSummary(name=task_name, **not_run_kwargs))
            self.write(prefix + _format_task_summary(task_summary) + '\n')



"""doit Linting Utilities."""

from pathlib import Path

from beartype import beartype
from beartype.typing import Iterable, List
from doit.tools import Interactive
from loguru import logger

from ..file_helpers import if_found_unlink
from .base import debug_task, echo
from .doit_globals import DG, DoitAction, DoitTask

# ----------------------------------------------------------------------------------------------------------------------
# Linting


@beartype
def _check_linting_errors(flake8_log_path: Path, ignore_errors: Iterable[str] = ()) -> None:  # noqa: CCR001
    """Check for errors reported in flake8 log file. Removes log file if no errors detected.

    Args:
        flake8_log_path: path to flake8 log file created with flag: `--output-file=flake8_log_path`
        ignore_errors: list of error codes to ignore (beyond the flake8 config settings). Default is None

    Raises:
        RuntimeError: if flake8 log file contains any text results

    """
    flake8_full_path = flake8_log_path.parent / f'{flake8_log_path.stem}-full{flake8_log_path.suffix}'
    log_contents = flake8_log_path.read_text().strip()
    review_info = f'. Review: {flake8_log_path}'
    if ignore_errors:
        # Backup the full list of errors
        flake8_full_path.write_text(log_contents)
        # Exclude the errors specified to be ignored by the user
        lines = log_contents.split('\n')
        lines = [
            line for line in lines
            if all(f': {error_code}' in line for error_code in ignore_errors)
        ]
        log_contents = '\n'.join(lines)
        flake8_log_path.write_text(log_contents)
        review_info = (
            f' even when ignoring {ignore_errors}.\nReview: {flake8_log_path}'
            f'\nNote: the full list linting errors are reported in {flake8_full_path}'
        )
    else:
        if_found_unlink(flake8_full_path)

    # Raise an exception if any errors were found. Remove the files if not
    if len(log_contents) > 0:
        raise RuntimeError(f'Found Linting Errors{review_info}')
    if_found_unlink(flake8_log_path)


@beartype
def _lint_python(
    lint_paths: List[Path], path_flake8: Path,
    ignore_errors: Iterable[str] = ('T100', 'T101'),
    xenon_args: str = '--max-absolute B --max-modules A --max-average A',
    diff_fail_under: int = 80, diff_branch: str = 'origin/main',
) -> List[DoitAction]:
    """Lint specified files creating summary log file of errors.

    Args:
        lint_paths: list of file and directory paths to lint
        path_flake8: path to flake8 configuration file
        ignore_errors: list of error codes to ignore (beyond the flake8 config settings). Default is to ignore Code Tags
        xenon_args: string arguments passed to xenon. Default is for most strict options available
        diff_fail_under: integer minimum test coverage. Default is 80
        diff_branch: string branch to compare against. Default is `origin/main`

    Returns:
        List[DoitAction]: doit task

    """
    # Flake8 appends to the log file. Ensure that an existing file is deleted so that Flake8 creates a fresh file
    run_m = 'poetry run python -m'
    flake8_log_path = DG.meta.path_project / 'flake8.log'
    flake8_flags = f'--config={path_flake8}  --output-file={flake8_log_path} --exit-zero'
    diff_params = f'--compare-branch={diff_branch} --fail-under={diff_fail_under}'
    diff_report = f'--html-report {DG.test.path_diff_lint_report}'
    return [
        (if_found_unlink, (flake8_log_path,)),
        Interactive(f'{run_m} flake8 {flake8_flags} ' + ' '.join(f'"{pth}"' for pth in lint_paths)),
        (_check_linting_errors, (flake8_log_path, ignore_errors)),
        # FIXME" Need to check if the branch is available first! Will fail on GHA
        Interactive(f'echo "poetry run diff-quality --violations=flake8 {diff_params} {diff_report}"'),
        Interactive(f'{run_m} xenon {DG.meta.pkg_name} {xenon_args}'),
    ]


@beartype
def _lint_non_python() -> List[DoitAction]:
    """Lint non-Python files such as JSON and YML/YAML.

    Returns:
        List[DoitAction]: doit task

    """
    actions = []
    pbs = DG.meta.paths_by_suffix
    paths_yaml = pbs.get('yml', []) + pbs.get('yaml', [])
    if paths_yaml:
        yamllint_args = '-d "{rules: {line-length: {max: 120}}}"'
        paths = ' '.join(f'"{pth}"' for pth in paths_yaml)
        actions.append(Interactive(f'poetry run yamllint {yamllint_args} {paths}'))

    # FYI: Use pre-commit instead
    # From: https://github.com/pre-commit/pre-commit-hooks/blob/0d261aaf84419c0c8fe70ff4a23f6a99655868de/
    #   lint: ./pre_commit_hooks/check_json.py
    #   format: ./pre_commit_hooks/pretty_format_json.py
    # > if paths_json := DG.meta.paths_by_suffix.get('json', []):
    # >     actions.extend(Interactive(f'poetry run jsonlint "{pth}"') for pth in paths_json)

    return actions


@beartype
def task_lint_python() -> DoitTask:
    """Lint all Python files and create summary of errors.

    Returns:
        DoitTask: doit task

    """
    actions = _lint_python(DG.lint.paths_py, path_flake8=DG.lint.path_flake8)
    return debug_task(actions)


@beartype
def task_lint_project() -> DoitTask:
    """Lint all project files that can be checked.

    Returns:
        DoitTask: doit task

    """
    actions = _lint_python(DG.lint.paths_py, path_flake8=DG.lint.path_flake8)
    actions.extend(_lint_non_python())
    return debug_task(actions)


@beartype
def task_lint_critical_only() -> DoitTask:
    """Suppress non-critical linting errors. Great for gating PRs/commits.

    Returns:
        DoitTask: doit task

    """
    actions = _lint_python(
        DG.lint.paths_py, path_flake8=DG.lint.path_flake8, ignore_errors=DG.lint.ignore_errors,
        xenon_args='--max-absolute C --max-modules A --max-average A',
    )
    actions.extend(_lint_non_python())
    return debug_task(actions)


@beartype
def task_radon_lint() -> DoitTask:
    """Lint project with Radon.

    See documentation: https://radon.readthedocs.io/en/latest/intro.html

    PLANNED: Simplify and choose one way of using Radon

    > Try Radon/Xenon, see [metrics](http://www.mccabe.com/iq_research_metrics.htm)

    Returns:
        DoitTask: doit task

    """
    actions: List[DoitAction] = []
    paths = ' '.join(map(Path.as_posix, DG.lint.paths_py))
    for args in ['mi --show --min B', 'cc --total-average --min B', 'hal', 'raw --summary']:
        actions.extend([
            (echo, (f'# Radon with args: {args}',)),
            # Could also create a file "--json --output-file .radon-{arg}.json" ("arg = args.split(' ')[0]")
            Interactive(f'poetry run radon {args} {paths}'),
        ])
    return debug_task(actions)


@beartype
def task_static_checks() -> DoitTask:
    """General static checkers (vulture, etc.).

    Returns:
        DoitTask: doit task

    """
    paths = ' '.join(map(Path.as_posix, DG.lint.paths_py))
    return debug_task([
        Interactive(f'poetry run vulture {paths} --min-confidence 70 --sort-by-size'),
    ])


@beartype
def task_security_checks() -> DoitTask:
    """Use linting tools to identify possible security vulnerabilities. Use `# nosec` to selectively override checks.

    Returns:
        DoitTask: doit task

    """
    return debug_task([
        Interactive(f'poetry run bandit --recursive {DG.meta.pkg_name}'),
        Interactive('poetry run nox --session check_safety'),
    ])


# ----------------------------------------------------------------------------------------------------------------------
# Formatting


@beartype
def _gen_format_actions(paths: str) -> List[str]:
    """Generate the list of format actions for Python files (isort, autopep8, etc.).

    # PLANNED: Considering splitting paths on commas if above some maximum character length?

    Args:
        paths: string list of one or more paths to pass to the cli tools

    Returns:
        DoitTask: doit task

    """
    run = 'poetry run'
    run_mod = f'{run} python -m'
    autoflake_args = (
        '--in-place --remove-all-unused-imports --remove-unused-variables --ignore-init-module-imports'
        ' --remove-duplicate-keys'
    )
    return [
        f'{run} pyupgrade {paths} --py38-plus --keep-runtime-typing',
        f'{run_mod} autoflake {paths} {autoflake_args}',
        f'{run_mod} autopep8 {paths} --in-place --aggressive',
        f'{run} pycln --quiet {paths}',
        f'{run} absolufy-imports {paths} --never',
        f'{run_mod} isort {paths} --settings-path "{DG.lint.path_isort}"',
        f'{run} add-trailing-comma {paths} --py36-plus --exit-zero-even-if-changed',
    ]


@beartype
def task_format_py() -> DoitTask:
    """Format a space-separate list of Python file(s). Particularly useful for pre-commit.

    ```sh
    poetry run format_py -p dodo.py tests/conftest.py
    ```

    Returns:
        DoitTask: doit task

    """
    return {
        'actions': _gen_format_actions('%(py_paths)s'),
        'pos_arg': 'py_paths',
        'verbosity': 2,
    }


# FIXME: Add bulk toml formatting to the auto_format task!
@beartype
def task_format_toml() -> DoitTask:
    """Format a space-separate list of TOML file(s) with `taplo`. Particularly useful for pre-commit.

    ```sh
    poetry run format_toml -p pyproject.toml .deepsource.toml
    ```

    Returns:
        DoitTask: doit task

    """
    return {
        # PLANNED: Could provide more hooks for configuring taplo options. See:
        #   https://taplo.tamasfe.dev/configuration/#formatting-options
        'actions': [
            # FIXME: capturing to /dev/null doesn't seem to work reliably
            Interactive(
                '(which taplo >> /dev/null && taplo format --options="indent_string=\'    \'" %(toml_paths)s) || true',
            ),
        ],
        'pos_arg': 'toml_paths',
        'verbosity': 2,
    }


@beartype
def task_auto_format() -> DoitTask:
    """Format code with isort, autopep8, and others.

    Additional snippets that may be useful, but aren't used by doit:

    ```sh
    poetry run isort --recursive --check --diff calcipy/ tests/
    ```

    Returns:
        DoitTask: doit task

    """
    @beartype
    def get_short_path(_pth: Path) -> Path:
        try:
            return _pth.relative_to(DG.meta.path_project)
        except Exception as exc:
            logger.warning(
                '{pth} is not relative to {rel_path}. Error: {exc}',
                pth=_pth, rel_path=DG.meta.path_project, exc=exc,
            )
            return _pth

    paths = ' '.join(f'"{get_short_path(pth)}"' for pth in DG.lint.paths_py)
    return debug_task(_gen_format_actions(paths))


# ----------------------------------------------------------------------------------------------------------------------
# General Static Analysis


@beartype
def task_pre_commit_hooks() -> DoitTask:
    """Run the pre-commit hooks  on all files.

    > Note: use `git commit` or `git push` with `--no-verify` if needed

    Returns:
        DoitTask: doit task

    """
    # Hooks should be installed for all types
    # https://github.com/pre-commit/pre-commit/blob/7858ad066f2dfd11453c2f7e25c8f055ba4de931/pre_commit/commands/install_uninstall.py#L103-L130
    return debug_task([
        Interactive('poetry run pre-commit install'),
        Interactive('poetry run pre-commit autoupdate'),
        Interactive('poetry run pre-commit run --all-files --hook-stage commit --hook-stage push'),
    ])



"""Custom Pytest Configuration.

To use the custom markers, create a file `tests/conftest.py` and add this import:

```python3
from calcipy.dev.conftest import pytest_configure  # noqa: F401
```

For HTML Reports, see: https://pypi.org/project/pytest-html/.

```python3
'''Custom PyTest-HTML Report Configuration.'''

from calcipy.dev.conftest import pytest_html_results_table_header  # noqa: F401
from calcipy.dev.conftest import pytest_html_results_table_row  # noqa: F401
from calcipy.dev.conftest import pytest_runtest_makereport  # noqa: F401
```

"""

from datetime import datetime

from beartype.typing import Any, Generator
from loguru import logger

_HAS_TEST_IMPORTS = False
try:
    import pytest
    from py.xml import html
    _HAS_TEST_IMPORTS = True
except ImportError:  # pragma: no cover
    ...

if _HAS_TEST_IMPORTS:
    @pytest.mark.optionalhook()
    def pytest_html_results_table_header(cells: Any) -> None:  # pragma: no cover
        """Modify results table in the pytest-html output.

        Args:
            cells: argument from pytest

        """
        cells.insert(1, html.th('Description'))
        cells.insert(1, html.th('Time', class_='sortable time', col='time'))

    @pytest.mark.optionalhook()
    def pytest_html_results_table_row(report: Any, cells: Any) -> None:  # pragma: no cover
        """Modify results table in the pytest-html output.

        Args:
            report: argument from pytest
            cells: argument from pytest

        """
        try:
            cells.insert(1, html.td(report.description))
            cells.insert(1, html.td(str(datetime.utcnow()), class_='col-time'))
        except AttributeError:
            ...  # The test suite likely failed

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_makereport(item: Any, call: Any) -> Generator:  # type: ignore[type-arg]  # pragma: no cover
        """Modify the pytest-html output.

        Args:
            item: argument from pytest
            call: argument from pytest

        Yields:
            outcome: required by pytest

        """
        try:
            logger.debug('call excinfo={excinfo}', excinfo=call.excinfo)
            outcome = yield
            report = outcome.get_result()
            report.description = str(item.function.__doc__)
            report.duration_formatter = '%H:%M:%S.%f'
        except AttributeError:
            ...  # The test suite likely failed

    def pytest_configure(config: Any) -> None:
        """Configure pytest with custom markers (SLOW, INTERACTIVE, and CURRENT).

        Args:
            config: pytest configuration object

        """
        config.addinivalue_line(
            'markers',
            'SLOW: tests that take a long time (>15s) and can be skipped with `-m "not SLOW"`',
        )
        config.addinivalue_line(
            'markers',
            'INTERACTIVE: tests that open a Chrome window and can be skipped with `-m "not INTERACTIVE"`',
        )
        config.addinivalue_line(
            'markers',
            'CURRENT: tests that are currently being developed. Useful for TDD with ptw `poetry run ptw -- -m CURRENT`',
        )



"""Find Files."""

from collections import defaultdict
from pathlib import Path

from beartype import beartype
from beartype.typing import Dict, List
from loguru import logger
from pre_commit.git import zsplit
from pre_commit.util import cmd_output


@beartype
def _get_all_files(*, cwd: Path) -> List[str]:
    """Get all files using git. Modified `pre_commit.git.get_all_files` to accept `cwd`.

    https://github.com/pre-commit/pre-commit/blob/488b1999f36cac62b6b0d9bc8eae99418ae5c226/pre_commit/git.py#L153

    Args:
        cwd: current working directory to pass to `subprocess.Popen`

    Returns:
        List[str]: list of all file paths relative to the `cwd`

    """
    return zsplit(cmd_output('git', 'ls-files', '-z', cwd=cwd)[1])  # type: ignore[no-any-return]


@beartype
def _filter_files(rel_filepaths: List[str], ignore_patterns: List[str]) -> List[str]:
    """Filter a list of string file paths with specified ignore patterns in glob syntax.

    Args:
        rel_filepaths: list of string file paths
        ignore_patterns: glob ignore patterns

    Returns:
        List[str]: list of all non-ignored file path names

    """
    if ignore_patterns:
        matches = []
        for pth in map(Path, rel_filepaths):
            matches.extend([pth.as_posix() for pat in ignore_patterns if pth.match(pat)][:1])
        return [rel for rel in rel_filepaths if rel not in matches]
    return rel_filepaths


@beartype
def find_project_files(path_project: Path, ignore_patterns: List[str]) -> List[Path]:
    """Find project files in git version control.

    > Note: uses the relative project directory and verifies that each file exists

    Args:
        path_project: Path to the project directory. Typically `DG.meta.path_project`
        ignore_patterns: glob ignore patterns

    Returns:
        Dict[str, List[Path]]: where keys are the suffix (without leading dot) and values the list of paths

    """
    file_paths = []
    rel_filepaths = _get_all_files(cwd=path_project)
    filtered_rel_files = _filter_files(rel_filepaths=rel_filepaths, ignore_patterns=ignore_patterns)
    for rel_file in filtered_rel_files:
        path_file = path_project / rel_file
        if path_file.is_file():
            file_paths.append(path_file)
        else:  # pragma: no cover
            logger.warning(f'Could not find {rel_file} in {path_project}')
    return file_paths


@beartype
def find_project_files_by_suffix(path_project: Path, ignore_patterns: List[str]) -> Dict[str, List[Path]]:
    """Find project files in git version control.

    > Note: uses the relative project directory and verifies that each file exists

    Args:
        path_project: Path to the project directory. Typically `DG.meta.path_project`
        ignore_patterns: glob ignore patterns

    Returns:
        Dict[str, List[Path]]: where keys are the suffix (without leading dot) and values the list of paths

    """
    file_lookup = defaultdict(list)
    for path_file in find_project_files(path_project, ignore_patterns):
        file_lookup[path_file.suffix.lstrip('.')].append(path_file)
    return file_lookup



"""nox-poetry configuration file.

[Useful snippets from docs](https://nox.thea.codes/en/stable/usage.html)

```sh
poetry run nox -l
poetry run nox --list-sessions

poetry run nox -s build_check-3.9 build_dist-3.9 check_safety-3.9
poetry run nox --session check_safety-3.9

poetry run nox --python 3.8

poetry run nox -k "not tests and not check_safety"
```

Useful nox snippets

```python3
# Example conditionally skipping a session
if not session.interactive:
    session.skip('Cannot run detect-secrets audit in non-interactive shell')

# Install pinned version
session.install('detect-secrets==1.0.3')

# Example capturing STDOUT into a file (could do the same for stderr)
path_stdout = Path('.stdout.txt').resolve()
with open(path_stdout, 'w') as out:
    session.run(*shlex.split('echo Hello World!'), stdout=out)
```

"""

import re
import shlex
from contextlib import suppress
from pathlib import Path
from urllib.parse import urlparse
from urllib.request import url2pathname

from beartype import beartype
from beartype.typing import Callable, Dict, List
from loguru import logger

from ..doit_tasks.doit_globals import DG, DoitAction, DoitTask
from ..doit_tasks.test import task_coverage, task_test
from ..file_helpers import if_found_unlink

_HAS_TEST_IMPORTS = False
try:
    from nox_poetry import session as nox_session
    from nox_poetry.poetry import DistributionFormat
    from nox_poetry.sessions import Session
    _HAS_TEST_IMPORTS = True
except ImportError:  # pragma: no cover
    ...

_DEV_KEY = '_dev'
"""Key for list of development dependencies."""

_PINS: Dict[str, List[str]] = {_DEV_KEY: []}
"""Hackish global to track dev-dependencies that are user-specified."""


@beartype
def pin_dev_dependencies(pins: List[str]) -> None:
    """Manually specify dependencies for development.

    TODO: Make this auto-resolve based on pyproject.toml

    Args:
        pins: list of dependencies (i.e. `['Cerberus=>1.3.4', 'freezegun']`)

    """
    _PINS[_DEV_KEY] = pins


if _HAS_TEST_IMPORTS:  # pragma: no cover  # noqa: C901
    def _run_str_cmd(session: Session, cmd_str: str) -> None:
        """Run a command string. Ensure that poetry is left-stripped.

        Args:
            session: nox_poetry Session
            cmd_str: string command to run

        """
        cmd_str = re.sub(r'^poetry run ', '', cmd_str)
        session.run(*shlex.split(cmd_str), stdout=True)

    def _install_calcipy_extras(session: Session, extras: List[str]) -> None:
        """Ensure calcipy extras are installed.

        Args:
            session: nox_poetry Session
            extras: list of extras to install

        """
        if DG.meta.pkg_name == 'calcipy':
            session.poetry.installroot(extras=extras)
        else:  # pragma: no cover
            session.install('.', f'calcipy[{",".join(extras)}]')

    def _install_pinned(session: Session, key: str) -> None:
        """Ensure user-pinned dependencies are installed.

        See [Issue #230](https://github.com/cjolowicz/nox-poetry/issues/230)

        Args:
            session: nox_poetry Session

        """
        for pin in _PINS.get(key, []):
            session.install(pin)

    def _run_func_cmd(action: DoitAction) -> None:
        """Run a python action.

        Args:
            action: doit python action

        Raises:
            RuntimeError: if a function action fails

        """
        # https://pydoit.org/tasks.html#python-action
        func, args, kwargs = [*list(action), {}][:3]
        result = func(args, **kwargs)
        if result not in [True, None] or not isinstance(result, (str, dict)):
            raise RuntimeError(f'Returned {result}. Failed to run task: {action}')

    def _run_doit_task(session: Session, task_fun: Callable[[], DoitTask]) -> None:
        """Run a DoitTask actions without using doit.

        Args:
            session: nox_poetry Session
            task_fun: function that returns a DoitTask

        Raises:
            NotImplementedError: if the action is of an unknown type

        """
        task = task_fun()
        for action in task['actions']:
            if isinstance(action, str):
                _run_str_cmd(session, action)
            elif getattr(action, 'action', None):
                _run_str_cmd(session, action.action)
            elif isinstance(action, (list, tuple)):
                _run_func_cmd(action)
            else:
                raise NotImplementedError(f'Unable to run {action} ({type(action)})')

    @nox_session(python=DG.test.pythons, reuse_venv=True)
    def tests(session: Session) -> None:
        """Run doit test task for specified python versions.

        Args:
            session: nox_poetry Session

        """
        _install_calcipy_extras(session, ['dev', 'test'])
        _install_pinned(session, key=_DEV_KEY)
        _run_doit_task(session, task_test)

    @nox_session(python=DG.test.pythons[-1:], reuse_venv=True)
    def coverage(session: Session) -> None:
        """Run doit test task for specified python versions.

        Args:
            session: nox_poetry Session

        """
        _install_calcipy_extras(session, ['dev', 'test'])
        _install_pinned(session, key=_DEV_KEY)
        _run_doit_task(session, task_coverage)

    @nox_session(python=DG.test.pythons[-1:], reuse_venv=False)
    def build_dist(session: Session) -> None:
        """Build the project files within a controlled environment for repeatability.

        Args:
            session: nox_poetry Session

        """
        if_found_unlink(DG.meta.path_project / 'dist')
        path_wheel = session.poetry.build_package()
        logger.info(f'Created wheel: {path_wheel}')
        # Install the wheel and check that imports without any of the optional dependencies
        session.install(path_wheel)
        session.run(*shlex.split('python scripts/check_imports.py'), stdout=True)

    @nox_session(python=DG.test.pythons[-1:], reuse_venv=True)
    def build_check(session: Session) -> None:
        """Check that the built output meets all checks.

        Args:
            session: nox_poetry Session

        """
        # Build sdist and fix return URI, which will have file://...#egg=calcipy
        sdist_uri = session.poetry.build_package(distribution_format=DistributionFormat.SDIST)
        path_sdist = Path(url2pathname(urlparse(sdist_uri).path))
        logger.debug(f'Fixed sdist URI ({sdist_uri}): {path_sdist}')
        # Check with pyroma
        session.install('pyroma', '--upgrade')
        # PLANNED: Troubleshoot why pyroma score is so low (6/10)
        session.run('pyroma', '--file', path_sdist.as_posix(), '--min=6', stdout=True)

    @nox_session(python=DG.test.pythons[-1:], reuse_venv=True)
    def check_safety(session: Session) -> None:
        """Check for known vulnerabilities with safety.

        Based on: https://github.com/pyupio/safety/issues/201#issuecomment-632627366

        Args:
            session: nox_poetry Session

        Raises:
            RuntimeError: if safety exited with errors, but not caught by session

        """
        session.install('safety', '--upgrade')
        path_report = Path('insecure_report.json').resolve()
        logger.info(f'Creating safety report: {path_report}')
        session.run(*shlex.split(f'safety check --full-report --cache --output {path_report} --json'), stdout=True)
        if path_report.read_text().strip() != '[]':
            raise RuntimeError(f'Found safety warnings in {path_report}')
        path_report.unlink()

    @nox_session(python=DG.test.pythons[-1:], reuse_venv=True)
    def check_security(session: Session) -> None:
        """More general checks for common security issues.

        Args:
            session: nox_poetry Session

        """
        session.install('semgrep', '--upgrade')  # Runs "safety" and other tools
        allow_py_rules = '--dangerously-allow-arbitrary-code-execution-from-rules'
        # TODO: Implement semgrep - what are a good ruleset to start with? Currently only a nox-session (no doit task)
        #   https://github.com/returntocorp/semgrep-rules/tree/develop/python
        #   https://awesomeopensource.com/project/returntocorp/semgrep-rules?categorypage=45
        configs = ' '.join([
            # See more at: https://semgrep.dev/explore
            '--config=p/ci',
            '--config=p/security-audit',
            '--config=r/python.airflow',
            '--config=r/python.attr',
            '--config=r/python.click',
            '--config=r/python.cryptography',
            '--config=r/python.distributed',
            '--config=r/python.docker',
            '--config=r/python.flask',
            '--config=r/python.jinja2',
            '--config=r/python.jwt',
            '--config=r/python.lang',
            '--config=r/python.pycryptodome',
            '--config=r/python.requests',
            '--config=r/python.security',
            '--config=r/python.sh',
            '--config=r/python.sqlalchemy',
            # dlukeomalley:unchecked-subprocess-call
            # dlukeomalley:use-assertEqual-for-equality
            # dlukeomalley:flask-set-cookie
            # clintgibler:no-exec
        ])
        session.run(*shlex.split(f'semgrep {DG.meta.pkg_name} {allow_py_rules} {configs}'), stdout=True)

    # TODO: https://github.com/tonybaloney/wily/blob/e72b7d95228bbe5538a072dc5d1186daa318bb03/src/wily/__main__.py#L261
    @nox_session(python=DG.test.pythons[-1:], reuse_venv=True)
    def check_wily(session: Session) -> None:
        """Run `wily` separately because of dependency constraints.

        Args:
            session: nox_poetry Session

        """
        session.install('wily', '--upgrade')
        logger.warning('FYI: Can only be run when all changes are checked in or stashed')
        # All possible metrics can be listed with: wily list-metrics
        # TODO: How are the additional metrics configured?
        # operators = ','.join([
        #     'raw.loc', 'raw.lloc', 'raw.sloc', 'maintainability.rank', 'maintainability.mi', 'cyclomatic.complexity',
        #     'halstead.h1', 'halstead.vocabulary', 'halstead.length', 'halstead.volume', 'halstead.difficulty', 'halstead.effort',
        # ])
        build_args = '--max-revisions 50'
        report_args = f'--output {DG.meta.path_project}/report.txt --message --number 50'
        with suppress(Exception):
            session.run(*shlex.split(f'wily build {DG.meta.pkg_name} {build_args}'), stdout=True)
            session.run(*shlex.split(f'wily report {DG.meta.pkg_name} {report_args}'), stdout=True)

            # FYI: Opens plotly graph in web browser and not necessary to run "wily" again
            # metric = 'raw.loc'  # 'cyclomatic.complexity' 'maintainability.mi' 'halstead.h1'
            # session.run(*shlex.split(f'wily graph {DG.meta.pkg_name} {metric}'), stdout=True)

        # TODO: Could use file archiver instead of git when the above fails?



"""doit Helpers.

Register all defaults doit tasks in a dodo.py file with the below snippet:

`from calcipy.doit_tasks import *  # noqa: F401,F403,H303 (Run 'doit list' to see tasks). # skipcq: PYL-W0614`

"""

__all__ = [  # noqa: F405
    'DOIT_CONFIG_RECOMMENDED',
    'TASKS_CI',
    'TASKS_LOCAL',
    # from .base
    'task_zip_release',
    # from .code_tags
    'task_collect_code_tags',
    # from .doc
    'task_cl_bump_pre',
    'task_cl_bump',
    'task_cl_write',
    'task_deploy_docs',
    'task_document',
    'task_open_docs',
    # from .lint
    'task_auto_format',
    'task_format_py',
    'task_format_toml',
    'task_lint_critical_only',
    'task_lint_project',
    'task_lint_python',
    'task_pre_commit_hooks',
    'task_radon_lint',
    'task_security_checks',
    'task_static_checks',
    # from .packaging
    'task_check_for_stale_packages',
    'task_check_license',
    'task_lock',
    'task_publish',
    # from .test
    'task_check_types',
    'task_coverage',
    'task_nox_test',
    'task_nox_coverage',
    'task_nox',
    'task_open_test_docs',
    'task_ptw_current',
    'task_ptw_ff',
    'task_ptw_marker',
    'task_ptw_not_interactive',
    'task_test_all',
    'task_test_keyword',
    'task_test_marker',
    'task_test',
]

from getpass import getuser

from .base import task_zip_release  # noqa: F401
from .code_tags import task_collect_code_tags  # noqa: F401
from .doc import *  # noqa: F401,F403,H303. lgtm [py/polluting-import]
from .lint import *  # noqa: F401,F403,H303. lgtm [py/polluting-import]
from .packaging import *  # noqa: F401,F403,H303. lgtm [py/polluting-import]
from .summary_reporter import SummaryReporter
from .test import *  # noqa: F401,F403,H303. lgtm [py/polluting-import]

TASKS_CI = [
    'nox_test',
    'security_checks',
]
"""More forgiving tasks to be run in CI."""

TASKS_LOCAL = [
    'collect_code_tags',
    'cl_write',
    'lock',
    'nox_coverage',
    'auto_format',
    'document',
    'check_for_stale_packages',
    'pre_commit_hooks',
    'lint_project',
    'static_checks',
    'security_checks',
    'check_types',
]
"""Full suite of tasks for local development."""

DOIT_CONFIG_RECOMMENDED = {
    'action_string_formatting': 'old',  # Required for keyword-based tasks
    'default_tasks': TASKS_CI if getuser().lower() == 'appveyor' else TASKS_LOCAL,
    'reporter': SummaryReporter,
}
"""doit Configuration Settings. Run with `poetry run doit`."""



"""doit Test Utilities."""

from beartype import beartype
from doit.tools import Interactive

from .base import debug_task, open_in_browser
from .doit_globals import DG, DoitTask

# ----------------------------------------------------------------------------------------------------------------------
# Manage Testing with Nox


@beartype
def _run_nox(args: str) -> DoitTask:
    """Run a nox command and fail if the interpreter is not found

    Returns:
        DoitTask: doit task

    """
    return debug_task([
        Interactive(f'poetry run nox --error-on-missing-interpreters {args}'),
    ])


@beartype
def task_nox() -> DoitTask:
    """Run the full nox test suite.

    > Note: some nox tasks are run in more-specific doit tasks, but this will run everything

    Returns:
        DoitTask: doit task

    """
    return _run_nox('')


@beartype
def task_nox_test() -> DoitTask:
    """Run the nox tests sessions.

    Returns:
        DoitTask: doit task

    """
    return _run_nox('--session tests')


@beartype
def task_nox_coverage() -> DoitTask:
    """Run the coverage session in nox.

    Returns:
        DoitTask: doit task

    """
    return _run_nox('--session coverage')


# ----------------------------------------------------------------------------------------------------------------------
# Manage Testing with pytest (Should be run from Nox)


@beartype
def task_test() -> DoitTask:
    """Run tests with Pytest and stop on the first failure.

    > Test are randomly ordered by default with pytest-randomly because that can help catch common errors
    > Tests can be re-run in the last order with `poetry run pytest --randomly-seed=last`

    > Tip: `--record-mode=rewrite` can be useful if working with `pytest-recording`

    Returns:
        DoitTask: doit task

    """
    return debug_task([
        Interactive(f'poetry run python -m pytest "{DG.test.path_tests}" {DG.test.args_pytest}'),
    ])


@beartype
def task_test_all() -> DoitTask:
    """Run all possible tests with Pytest even if one or more failures.

    Returns:
        DoitTask: doit task

    """
    return debug_task([
        Interactive(f'poetry run python -m pytest "{DG.test.path_tests}" --ff -vv'),
    ])


@beartype
def task_test_marker() -> DoitTask:
    """Specify a marker to run a subset of tests.

    Example: `doit run test_marker -m "not MARKER"` or `doit run test_marker -m "MARKER"`

    Returns:
        DoitTask: doit task

    """
    task = debug_task(
        [Interactive(f'poetry run python -m pytest "{DG.test.path_tests}" {DG.test.args_pytest} -m "%(marker)s"')],
    )
    task['params'] = [{
        'name': 'marker', 'short': 'm', 'long': 'marker', 'default': '',
        'help': (
            'Runs test with specified marker logic\nSee: '
            'https://docs.pytest.org/en/latest/example/markers.html?highlight=-m'
        ),
    }]
    return task


@beartype
def task_test_keyword() -> DoitTask:
    """Specify a keyword to run a subset of tests.

    Example: `doit run test_keyword -k "KEYWORD"`

    Returns:
        DoitTask: doit task

    """
    return {
        'actions': [
            Interactive(f'poetry run python -m pytest "{DG.test.path_tests}" {DG.test.args_pytest} -k "%(keyword)s"'),
        ],
        'params': [{
            'name': 'keyword', 'short': 'k', 'long': 'keyword', 'default': '',
            'help': (
                'Runs only tests that match the string pattern\nSee: '
                'https://docs.pytest.org/en/latest/usage.html#specifying-tests-selecting-tests'
            ),
        }],
        'verbosity': 2,
    }


@beartype
def task_coverage() -> DoitTask:
    """Run pytest and create coverage and test reports.

    Returns:
        DoitTask: doit task

    """
    path_tests = DG.test.path_tests
    cov_dir = DG.test.path_coverage_index.parent
    test_html = f'--html="{DG.test.path_test_report}" --self-contained-html'
    diff_html = f'--html-report {DG.test.path_diff_test_report}'
    return debug_task([
        Interactive(
            f'poetry run coverage run --source={DG.meta.pkg_name} --module'
            f' pytest "{path_tests}" {DG.test.args_pytest} {test_html}',
        ),
        'poetry run python -m coverage report --show-missing',
        f'poetry run python -m coverage html --directory={cov_dir}',
        'poetry run python -m coverage json',  # Create coverage.json file for "_write_coverage_to_md"
        'poetry run python -m coverage xml',
        Interactive(f'poetry run diff-cover coverage.xml {DG.test.args_diff} {diff_html}'),
    ])


# ----------------------------------------------------------------------------------------------------------------------
# Other Test Tools (MyPy, etc.)


@beartype
def task_check_types() -> DoitTask:
    """Run type annotation checks.

    Returns:
        DoitTask: doit task

    """
    return debug_task([
        Interactive(f'poetry run mypy {DG.meta.pkg_name} --show-error-codes'),
    ])


# ----------------------------------------------------------------------------------------------------------------------
# Test Output Interaction


@beartype
def task_open_test_docs() -> DoitTask:
    """Open the test and coverage files in default browser.

    Returns:
        DoitTask: doit task

    """
    paths = [
        DG.test.path_test_report,
        DG.test.path_diff_test_report,
        DG.test.path_diff_lint_report,
        DG.test.path_coverage_index,
        DG.test.path_mypy_index,
    ]
    return debug_task([(open_in_browser, (pth,)) for pth in paths if pth.is_file()])


# ----------------------------------------------------------------------------------------------------------------------
# Implement long running ptw tasks


@beartype
def ptw_task(cli_args: str) -> DoitTask:
    """Return doit Interactive `ptw` task.

    Args:
        cli_args: string CLI args to pass to `ptw`

    Returns:
        DoitTask: doit task

    """
    return {
        'actions': [Interactive(f'poetry run ptw "{DG.meta.path_project}" {cli_args}')],
        'verbosity': 2,
    }


@beartype
def task_ptw_not_interactive() -> DoitTask:
    """Run pytest watch for failed first and skip the INTERACTIVE marker.

    kwargs: `-m 'not INTERACTIVE' -vvv`

    Returns:
        DoitTask: doit task

    """
    return ptw_task('-m "not INTERACTIVE" -vvv')


@beartype
def task_ptw_ff() -> DoitTask:
    """Run pytest watch for failed first and skip the INTERACTIVE marker.

    kwargs: `--last-failed --new-first -m 'not INTERACTIVE' -vv`

    Returns:
        DoitTask: doit task

    """
    return ptw_task('--last-failed --new-first -m "not INTERACTIVE" -vv')


@beartype
def task_ptw_current() -> DoitTask:
    """Run pytest watch for only tests with the CURRENT marker.

    kwargs: `-m 'CURRENT' -vv`

    Returns:
        DoitTask: doit task

    """
    return ptw_task('-m "CURRENT" -vv')


@beartype
def task_ptw_marker() -> DoitTask:
    """Specify a marker to run a subset of tests in Interactive `ptw` task.

    Example: `doit run ptw_marker -m "not MARKER"` or `doit run ptw_marker -m "MARKER"`

    Returns:
        DoitTask: doit task

    """
    task = ptw_task('-vvv -m "%(marker)s"')
    task['params'] = [{
        'name': 'marker', 'short': 'm', 'long': 'marker', 'default': '',
        'help': (
            'Runs test with specified marker logic\nSee: '
            'https://docs.pytest.org/en/latest/example/markers.html?highlight=-m'
        ),
    }]
    return task



"""Collect code tags and output for review in a single location."""

from beartype import beartype

from ..code_tag_collector import write_code_tag_file
from ..file_helpers import get_relative
from .base import debug_task
from .doit_globals import DG, DoitTask


@beartype
def task_collect_code_tags() -> DoitTask:
    """Create a summary file with all of the found code tags.

    Returns:
        DoitTask: doit task

    """
    # Filter out any files that are auto-generated and would have duplicate code tags, such as the documentation
    user_files = [pth for pth in DG.meta.paths if not get_relative(pth, DG.doc.auto_doc_path)]
    kwargs = {
        'path_tag_summary': DG.tags.path_code_tag_summary,
        'paths_source': user_files,
        'base_dir': DG.meta.path_project,
        'regex_compiled': DG.tags.compile_issue_regex(),
        'header': f'# Task Summary\n\nAuto-Generated by `{DG.meta.pkg_name}`',
        'tag_order': DG.tags.tags,
    }
    return debug_task([(write_code_tag_file, (), kwargs)])



"""Calcipy Command Line Interface."""

# PLANNED: Take a look at a few useful issues for cement:
# FAQ: The right decision to layout your Cement 3 CLI app (and test)
# 	https://github.com/datafolklabs/cement/issues/602



"""Global Variables for doit."""

import inspect
import re
import warnings
from functools import partial
from pathlib import Path

import doit
import tomli
from attrs import field, mutable
from attrs_strict import type_validator
from beartype import beartype
from beartype.typing import Any, Callable, Dict, Iterable, List, Optional, Pattern, Set, Tuple, Union
from doit.action import BaseAction
from doit.task import Task
from loguru import logger

from ..code_tag_collector import CODE_TAG_RE, COMMON_CODE_TAGS
from ..file_helpers import _MKDOCS_CONFIG_NAME, _read_yaml_file, get_doc_dir
from ..file_search import find_project_files, find_project_files_by_suffix
from ..log_helpers import log_fun

_DOIT_TASK_IMPORT_ERROR = 'User must install the optional calcipy extra "dev" to utilize "doit_tasks"'
"""Standard error message when an optional import is not available. Raise with RuntimeError."""

_DoitCallableArgs = Iterable[Union[str, float, int, Path, Dict[str, Any]]]
"""Type: legal types that can be passed to a Python callable for doit actions."""

DoitAction = Union[str, BaseAction, Tuple[Callable, _DoitCallableArgs]]  # type: ignore[type-arg]
"""Type: individual doit action."""

DoitTask = Union[Task, Dict[str, DoitAction]]
"""Type: full doit task."""


@beartype
def _make_full_path(raw: Union[Path, str], path_base: Path) -> Path:
    """Return a full path by determining if the source path is an absolute path. If not combines with base path.

    Args:
        raw: (string or Path) relative or absolute path
        path_base: base directory to use if the raw path is not absolute

    Returns:
        Path: absolute path

    """
    return Path(raw) if Path(raw).is_absolute() else path_base / raw


@beartype
def _member_filter(member: Any, instance_type: Any) -> bool:
    """Return True if the member matches the filters.

    Args:
        member: class data- or method-member
        instance_type: optional instance type

    Returns:
        bool: True if the member matches the applied filter

    """
    return instance_type is None or isinstance(member, instance_type)


@mutable(kw_only=True)
class _PathAttrBase:  # noqa: H601

    path_project: Path = field(validator=type_validator())
    """Path to the package directory."""

    def __attrs_post_init__(self) -> None:
        """Initialize full paths with the package base directory if necessary.

        Raises:
            RuntimeError: if any paths are None

        """
        if self.path_project is None:
            raise RuntimeError('Missing keyword argument "path_project"')  # pragma: no cover
        self._resolve_class_paths(self.path_project)
        self._verify_initialized_paths()

    @beartype
    def _get_members(self, prefix: Optional[str], **kwargs: Any) -> List[Tuple[str, Any]]:
        """Return the members that match the parameters.

        Example to return all methods that start with `do_`: `self._get_members(instance_type=Callable, prefix='do_')`

        Args:
            prefix: optional string prefix to check starts with
            kwargs: keyword arguments passed to `_member_filter`

        Returns:
            List[Tuple[str, Any]]: filtered members from the class

        """
        members = inspect.getmembers(self, predicate=partial(_member_filter, **kwargs))
        return [
            (name, member) for (name, member) in members
            if not prefix or name.startswith(prefix)
        ]

    @beartype
    def _resolve_class_paths(self, base_path: Path) -> None:
        """Resolve all partial paths with the specified base path.

        WARN: will mutate the class attribute

        Args:
            base_path: base path to apply to all found relative paths

        """
        for name, path_raw in self._get_members(instance_type=type(Path()), prefix=None):
            if not path_raw.is_absolute():  # type: ignore[attr-defined]
                setattr(self, name, base_path / path_raw)  # type: ignore[operator]
                logger.debug(f'Mutated: self.{name}={path_raw} (now: {getattr(self, name)})')

    @beartype
    def _verify_initialized_paths(self) -> None:
        """Verify that all paths are not None.

        WARN: will not raise on error the class attribute

        Raises:
            RuntimeError: if any paths are None

        """
        missing = [name for name, _m in self._get_members(instance_type=type(None), prefix='path_')]
        if missing:
            kwargs = ', '.join(missing)
            raise RuntimeError(f'Missing keyword arguments for: {kwargs}')


@mutable(kw_only=True)
class PackageMeta(_PathAttrBase):  # noqa: H601
    """Package Meta-Information."""

    path_toml: Path = field(validator=type_validator(), default=Path('pyproject.toml'))
    """Relative path to the poetry toml file."""

    ignore_patterns: List[str] = field(validator=type_validator(), factory=list)
    """List of glob patterns to ignore from all analysis."""

    paths: List[Path] = field(validator=type_validator(), init=False)
    """Paths to all tracked files that were not ignored with specified patterns `find_project_files`."""

    paths_by_suffix: Dict[str, List[Path]] = field(validator=type_validator(), init=False)
    """Paths to all tracked files that were not ignored with specified patterns `find_project_files_by_suffix`."""

    pkg_name: str = field(validator=type_validator(), init=False)
    """Package string name."""

    pkg_version: str = field(validator=type_validator(), init=False)
    """Package version."""

    def __attrs_post_init__(self) -> None:
        """Finish initializing class attributes.

        Raises:
            RuntimeError: if the toml package is not available
            FileNotFoundError: if the toml could not be located

        """
        super().__attrs_post_init__()

        # Note: toml is an optional dependency required only when using the `doit_tasks` in development
        if tomli is None:  # pragma: no cover
            raise RuntimeError(_DOIT_TASK_IMPORT_ERROR)

        if not self.path_toml.is_file():
            raise RuntimeError(f'Check "{self.path_project}". Could not find: "{self.path_toml}"')

        poetry_config = tomli.loads(self.path_toml.read_text())['tool']['poetry']
        self.pkg_name = poetry_config['name']
        self.pkg_version = poetry_config['version']

        if '-' in self.pkg_name:  # pragma: no cover
            warnings.warn(f'Replace dashes in name with underscores ({self.pkg_name}) in {self.path_toml}')

        self.paths = find_project_files(self.path_project, self.ignore_patterns)
        self.paths_by_suffix = find_project_files_by_suffix(self.path_project, self.ignore_patterns)

    @beartype
    def __shorted_path_list(self) -> Set[str]:  # pragma: no cover
        """Shorten the list of directories common to the specified paths.

        > Not currently needed, but could be useful

        Returns:
            Set[str]: set of most common top-level directories relative to the project dir

        """
        return {
            pth.parent.relative_to(self.path_project).as_posix()
            for pth in self.paths
        }  # type: ignore[attr-defined]


_DEF_IGNORE_LIST = [
    'AAA01',  # AAA01 / act block in pytest
    'C901',  # C901 / complexity from "max-complexity = 10"
    'D417',  # D417 / missing arg descriptors
    'DAR101', 'DAR201', 'DAR401',  # https://pypi.org/project/darglint/ (Scroll to error codes)
    'DUO106',  # DUO106 / insecure use of os
    'E800',  # E800 / Commented out code
    'G001',  # G001 / logging format for un-indexed parameters
    'H601',  # H601 / class with low cohesion
    'P101', 'P103',  # P101,P103 / format string
    'PD013',
    'S101',  # S101 / assert
    'S605', 'S607',  # S605,S607 / os.popen(...)
    'T100', 'T101', 'T103',  # T100,T101,T103 / fixme and todo comments
]
"""Default list of excluded flake8 rules for the pre-commit check (additional to .flake8)."""


@mutable(kw_only=True)
class LintConfig(_PathAttrBase):  # noqa: H601
    """Lint Config."""

    path_flake8: Union[Path, str] = field(validator=type_validator(), default=Path('.flake8'))
    """Relative path to the flake8 configuration file. Default is ".flake8" created by calcipy_template."""

    path_isort: Union[Path, str] = field(validator=type_validator(), default=Path('pyproject.toml'))
    """Relative path to the isort configuration file. Default is "pyproject.toml" created by calcipy_template."""

    ignore_errors: List[str] = field(validator=type_validator(), factory=lambda: _DEF_IGNORE_LIST)
    """List of additional excluded flake8 rules for the pre-commit check."""

    paths_py: List[Path] = field(validator=type_validator(), init=False)
    """Paths to the Python files used when linting. Created with `find_project_files_by_suffix`."""

    def __attrs_post_init__(self) -> None:
        """Finish initializing class attributes."""
        super().__attrs_post_init__()
        self.path_flake8 = _make_full_path(self.path_flake8, self.path_project)
        self.path_isort = _make_full_path(self.path_isort, self.path_project)
        self.paths_py = DG.meta.paths_by_suffix.get('py', [])


@mutable(kw_only=True)
class TestingConfig(_PathAttrBase):  # noqa: H601  # pylint: disable=too-many-instance-attributes
    """Test Config."""

    pythons: List[str] = field(validator=type_validator(), factory=lambda: ['3.8', '3.9'])
    """Python versions to test against. Default is `['3.8', '3.9']`."""

    path_out: Union[Path, str] = field(validator=type_validator(), default=Path('releases/tests'))
    """Relative path to the report output directory. Default is `releases/tests`."""

    path_tests: Union[Path, str] = field(validator=type_validator(), default=Path('tests'))
    """Relative path to the tests directory. Default is `tests`."""

    args_pytest: str = field(
        validator=type_validator(),
        default='--exitfirst --showlocals --failed-first --new-first --verbose --doctest-modules',
    )
    """Default arguments to Pytest. In short form, the defaults are `-x -l --ff --nf -vv`."""

    args_diff: str = field(validator=type_validator(), default='--fail-under=65 --compare-branch=origin/main')
    """Default arguments to diff-cover."""

    path_test_report: Path = field(validator=type_validator(), init=False)
    """Path to the self-contained test HTML report."""

    path_diff_test_report: Path = field(validator=type_validator(), init=False)
    """Path to the self-contained diff-test HTML report."""

    path_diff_lint_report: Path = field(validator=type_validator(), init=False)
    """Path to the self-contained diff-lint HTML report."""

    path_coverage_index: Path = field(validator=type_validator(), init=False)
    """Path to the coverage HTML index file within the report directory."""

    path_mypy_index: Path = field(validator=type_validator(), init=False)
    """Path to the mypy HTML index file within the report directory."""

    def __attrs_post_init__(self) -> None:
        """Finish initializing class attributes."""
        super().__attrs_post_init__()
        self.path_out = _make_full_path(self.path_out, self.path_project)
        self.path_tests = _make_full_path(self.path_tests, self.path_project)
        self.path_out.mkdir(exist_ok=True, parents=True)
        # Configure the paths to the report HTML and coverage HTML files
        self.path_test_report = self.path_out / 'test_report.html'
        self.path_diff_test_report = self.path_out / 'diff_test_report.html'
        self.path_diff_lint_report = self.path_out / 'diff_lint_report.html'
        self.path_coverage_index = self.path_out / 'cov_html/index.html'
        self.path_mypy_index = self.path_out / 'mypy_html/index.html'


@mutable(kw_only=True)
class CodeTagConfig(_PathAttrBase):  # noqa: H601
    """Code Tag Config."""

    doc_sub_dir: Path = field(validator=type_validator(), default=Path('docs/docs'))
    """Relative path to the source documentation directory."""

    code_tag_summary_filename: str = field(validator=type_validator(), default='CODE_TAG_SUMMARY.md')
    """Name of the code tag summary file."""

    tags: List[str] = field(validator=type_validator(), factory=lambda: COMMON_CODE_TAGS)
    """List of ordered tag names to match."""

    re_raw: str = field(validator=type_validator(), default=CODE_TAG_RE)
    """string regular expression that contains `{tag}`."""

    path_code_tag_summary: Path = field(validator=type_validator(), init=False)
    """Path to the code tag summary file. Uses `code_tag_summary_filename`."""

    def __attrs_post_init__(self) -> None:
        """Finish initializing class attributes."""
        super().__attrs_post_init__()
        # Configure full path to the code tag summary file
        self.path_code_tag_summary = self.path_project / self.doc_sub_dir / self.code_tag_summary_filename

    @beartype
    def compile_issue_regex(self) -> Pattern[str]:
        """Compile the regex for the specified raw regular expression string and tags.

        Returns:
            Pattern[str]: compiled regular expression to match all of the specified tags

        """
        return re.compile(self.re_raw.format(tag='|'.join(self.tags)))


@mutable(kw_only=True)
class DocConfig(_PathAttrBase):  # noqa: H601
    """Documentation Config."""

    doc_sub_dir: Path = field(validator=type_validator(), default=Path('docs/docs'))
    """Relative path to the source documentation directory."""

    auto_doc_path: Optional[Path] = field(validator=type_validator(), default=None)
    """Auto-calculated based on `self.doc_sub_dir`."""

    handler_lookup: Optional[
        Dict[str, Callable[[str, Path], List[str]]]
    ] = field(validator=type_validator(), default=None)
    """Lookup dictionary for autoformatted sections of the project's markdown files."""

    path_out: Path = field(validator=type_validator(), init=False)
    """The documentation output directory. Specified in `mkdocs.yml`."""

    paths_md: List[Path] = field(validator=type_validator(), init=False)
    """Paths to Markdown files used when documenting. Created with `find_project_files_by_suffix`."""

    def __attrs_post_init__(self) -> None:
        """Finish initializing class attributes."""
        super().__attrs_post_init__()
        mkdocs_config = _read_yaml_file(self.path_project / _MKDOCS_CONFIG_NAME)
        self.path_out = Path(mkdocs_config.get('site_dir', 'releases/site'))
        self.path_out = _make_full_path(self.path_out, self.path_project)
        self.path_out.mkdir(exist_ok=True, parents=True)
        self.paths_md = DG.meta.paths_by_suffix.get('md', [])
        self.auto_doc_path = self.doc_sub_dir.parent / 'modules'


@mutable(kw_only=True)
class DoitGlobals:  # noqa: H601  # pylint: disable=too-many-instance-attributes
    """Global Variables for doit."""

    calcipy_dir: Path = field(validator=type_validator(), init=False, default=Path(__file__).resolve().parents[1])
    """The calcipy directory (likely within `.venv`)."""

    meta: PackageMeta = field(validator=type_validator(), init=False)
    """Package Meta-Information."""

    lint: LintConfig = field(validator=type_validator(), init=False)
    """Lint Config."""

    test: TestingConfig = field(validator=type_validator(), init=False)
    """Test Config."""

    tags: CodeTagConfig = field(validator=type_validator(), init=False)
    """Documentation Config."""

    doc: DocConfig = field(validator=type_validator(), init=False)
    """Documentation Config."""

    _is_set: bool = field(validator=type_validator(), default=False)
    """Internal flag to check if already set."""

    @log_fun
    def set_paths(
        self, *, path_project: Optional[Path] = None,
    ) -> None:
        """Configure `DoitGlobals` based on project directory.

        Args:
            path_project: optional project base directory Path. Defaults to the current working directory

        Raises:
            RuntimeError: if `set_paths` is called twice. This can cause unexpected behavior, so if
                absolutely necessary, set `DG._is_set = False`, then try `DG.set_paths()`

        """
        if self._is_set:
            raise RuntimeError('DoitGlobals has already been configured and cannot be reconfigured')

        logger.info(f'Setting DG path: {path_project}', path_project=path_project, cwd=Path.cwd())
        path_project = path_project or Path.cwd()

        # Read the optional toml configuration
        # > Note: could allow LintConfig/.../DocConfig kwargs to be set in toml, but may be difficult to maintain
        data = (path_project / 'pyproject.toml').read_text()
        calcipy_config = tomli.loads(data).get('tool', {}).get('calcipy', {})

        self._set_meta(path_project, calcipy_config)
        self._set_submodules(calcipy_config)

        self._is_set = True

    @beartype
    def _set_meta(self, path_project: Path, calcipy_config: Dict[str, Any]) -> None:
        """Initialize the meta submodules.

        Args:
            path_project: project base directory Path
            calcipy_config: custom calcipy configuration from toml file

        """
        ignore_patterns = calcipy_config.get('ignore_patterns', [])
        self.meta = PackageMeta(path_project=path_project, ignore_patterns=ignore_patterns)

    @beartype
    def _set_submodules(self, calcipy_config: Dict[str, Any]) -> None:
        """Initialize the rest of the submodules.

        Args:
            calcipy_config: custom calcipy configuration from toml file

        Raises:
            RuntimeError: if problems in formatting of the toml file

        """
        # Configure global options
        section_keys = ['lint', 'test', 'code_tag', 'doc']
        supported_keys = section_keys + ['ignore_patterns']
        unexpected_keys = [key for key in calcipy_config if key not in supported_keys]
        if unexpected_keys:
            raise RuntimeError(f'Found unexpected key(s) {unexpected_keys} (i.e. not in {supported_keys})')

        # Parse the Copier file for configuration information
        doc_sub_dir = get_doc_dir(self.meta.path_project) / 'docs'  # Note: subdirectory is important
        doc_sub_dir.mkdir(exist_ok=True, parents=True)

        # Configure submodules
        meta_kwargs = {'path_project': self.meta.path_project}
        lint_k, test_k, code_k, doc_k = (calcipy_config.get(key, {}) for key in section_keys)
        self.lint = LintConfig(**meta_kwargs, **lint_k)  # type: ignore[arg-type]
        self.test = TestingConfig(**meta_kwargs, **test_k)  # type: ignore[arg-type]
        self.tags = CodeTagConfig(**meta_kwargs, doc_sub_dir=doc_sub_dir, **code_k)  # type: ignore[arg-type]
        self.doc = DocConfig(**meta_kwargs, doc_sub_dir=doc_sub_dir, **doc_k)  # type: ignore[arg-type]


DG = DoitGlobals()
"""Global doit Globals class used to manage global variables."""

_WORK_DIR = doit.get_initial_workdir()
"""Work directory identified by doit."""

DG.set_paths(path_project=(Path(_WORK_DIR) if _WORK_DIR else Path.cwd()).resolve())



"""Dotted dictionary for consistent interface."""

from beartype import beartype
from beartype.typing import Any, Dict, Union
from box import Box

DDICT_TYPE = Union[Dict[str, Any], Box]
"""Return type from `ddict()`."""


@beartype
def ddict(**kwargs: Dict[str, Any]) -> DDICT_TYPE:
    """Return a dotted dictionary that can also be accessed normally.

    Currently uses `python-box` because there is a more recent release, but could also use `munch`

    Other variations are no longer supported, such as `bunch` and `ddict` among others

    Args:
        kwargs: keyword arguments formatted into dictionary

    Returns:
        DDICT_TYPE: dotted dictionary

    """
    return Box(kwargs)



"""General doit Utilities."""

import shutil
import webbrowser
from collections import defaultdict
from pathlib import Path

from beartype import beartype
from beartype.typing import Iterable, List
from doit.task import Task
from loguru import logger
from sty import fg

from ..file_helpers import if_found_unlink
from .doit_globals import DG, DoitAction, DoitTask


@beartype
def _show_cmd(task: Task) -> str:
    """For debugging, log the full command to the console.

    Args:
        task: doit Task

    Returns:
        str: describing the sequence of actions

    """
    def clean_action(action: str) -> str:
        action = action.replace(f'{DG.meta.path_project.as_posix()}/', '')
        if 'password' in action:
            action = action.split('password')[0] + '...password-redacted...'
        return action  # noqa: R504

    indent = '\n\t'
    actions = indent.join(map(clean_action, map(str, task.actions)))
    return f'{task.name} > [{indent}{fg.grey}{actions}{fg.rs}\n]\n'


@beartype
def debug_task(actions: Iterable[DoitAction], verbosity: int = 2) -> DoitTask:
    """Activate verbose logging for the specified actions.

    Args:
        actions: list of doit actions
        verbosity: 2 is maximum, while 0 is deactivated. Default is 2

    Returns:
        DoitTask: doit task

    """
    task: DoitTask = defaultdict(list)
    task['actions'] = actions
    task['title'] = _show_cmd
    task['verbosity'] = verbosity
    logger.debug('Created task. See extras', task=f'{task}')
    return task


@beartype
def echo(msg: str) -> None:
    """Wrap the system print command.

    Args:
        msg: string to write to STDOUT

    """
    print(msg)  # noqa: T001  # pragma: no cover


@beartype
def write_text(path_file: Path, text: str) -> None:
    """path_file.write_text wrapper for doit.

    Args:
        path_file: Path to the file
        text: string to write to file

    """
    path_file.parent.mkdir(exist_ok=True, parents=True)
    path_file.write_text(text)  # pragma: no cover


@beartype
def open_in_browser(path_file: Path) -> None:
    """Open the path in the default web browser.

    Args:
        path_file: Path to file

    """
    webbrowser.open(Path(path_file).as_uri())  # pragma: no cover


@beartype
def _make_archive(path_dir: Path, archive_format: str = 'zip') -> None:
    """Make archive.

    Args:
        path_dir: directory to zip
        archive_format: default is `zip`. See `shutil.make_archive` for supported types

    """
    shutil.make_archive(path_dir, archive_format, path_dir)


@beartype
def _zip_release() -> List[DoitAction]:
    """Zip up important information in the releases directory.

    > WARN: zip paths must be in the AppVeyor.yml to be collected

    Returns:
        DoitTask: doit task

    """
    actions = []
    for path_dir in [DG.test.path_out, DG.doc.path_out]:
        if path_dir.is_dir():
            path_zip = path_dir.with_suffix('.zip')
            actions.extend([
                (if_found_unlink, (path_zip,)),
                (_make_archive, (path_dir,)),
                (echo, (f'Created: {path_zip}',)),
            ])
    return actions


@beartype
def task_zip_release() -> DoitTask:
    """Zip up important information in the releases directory.

    Returns:
        DoitTask: doit task

    """
    return debug_task(_zip_release())



"""Version Utility."""

from beartype import beartype
from beartype.typing import Iterable
from cement.utils.version import get_version as cement_get_version

from ... import __version__

_VERSION = __version__.split('.')[:3] + ['beta', '0']  # Expects exactly 5 values


@beartype
def get_version(version: Iterable[str] = _VERSION) -> str:
    """Format the version for printing."""
    return cement_get_version(version)
